#!/usr/bin/perl -w

package B::ToXML;
use strict;

use B qw( walkoptree_slow svref_2object walkoptree_exec main_root );
use B::Asmdata qw( @specialsv_name );

# don't load HTML::Parser in HTML::Entities
BEGIN {
	$INC{'HTML/Parser.pm'} = 1;
}
use HTML::Entities ();
*encode = \&HTML::Entities::encode;

sub compile {
	# XXX - must filter out things printed from B::walkoptree_exec()
	if (@_ and $_[0] =~ /^-o/) {
		(my $file = shift) =~ s/^-o//;
		open(OUTPUT, "> $file") or die "Cannot open $file: $!";
		select OUTPUT;
	}
	my @funcs = @_;

	if (@funcs) {
		return sub {
			print qq|<functions>\n| if @funcs > 1;
			foreach my $func (@funcs) {

				my $pack;
				if ($func =~ s/^(.+):://) {
					$pack = $1;
				}
				$pack ||= 'main';
				warn "Using ($pack) for <$func>\n";
				
				my $funcref;
				next unless ($funcref = UNIVERSAL::can($pack, $func));

				print qq|<function name="${pack}::$func">\n|;
				toxml($funcref, 1);
				print "</function>\n";
			}
			print qq|</functions>| if @funcs > 1;
		};
	} else {
		return \&toxml;
	}
}

sub toxml {
	my $op;
	if (my $sub = shift) {
		$op = svref_2object($sub)->ROOT;
	} else {
		$op = main_root();
	}

	$op->toxml( shift );
	#	walkoptree_slow($op, 'toxml', shift);
}

sub B::OP::toxml {
	my ($op, $level) = @_;
	$level ||= 0;

	print indent($level), "<op ", B::ToXML::get_attrs($op), "/>\n";
}

sub B::UNOP::toxml {
	my ($op, $level) = @_;
	$level ||= 0;

	print indent($level), "<unop ", B::ToXML::get_attrs($op), ">\n";

	# if this is a null (optimized away), it may have multiple kids
	my $kid = $op->first;
	do {
		$kid->toxml( $level + 1 );
	} while ( $kid = $kid->sibling and $$kid );

	print indent($level), "</unop>\n";
}

sub B::BINOP::toxml {
	my ($op, $level) = @_;
	$level ||= 0;

	print indent($level), "<binop ", B::ToXML::get_attrs($op), ">\n";

	$op->first->toxml( $level + 1 );
	$op->last->toxml( $level + 1 );

	print indent($level), "</binop>\n";
}

sub B::LISTOP::toxml {
	my ($op, $level) = @_;
	$level ||= 0;

	print indent($level), "<listop ", B::ToXML::get_attrs($op), ">\n";

	my $kid = $op->first;
	do {
		$kid->toxml( $level + 1 );
	} while ( $kid = $kid->sibling and $$kid );

	print indent($level), "</listop>\n";
}

sub B::SVOP::toxml {
	my ($op, $level) = @_;
	$level ||= 0;

	print indent($level), "<svop ", B::ToXML::get_attrs($op), ">\n";
	$op->sv->toxml( $level + 1 );
	print indent($level), "</svop>\n";
}

sub B::PADOP::toxml {
	my ($op, $level) = @_;

	print indent($level), "<padop ", 
		B::ToXML::get_attrs($op, 'padix', 'precomp'), "/>\n";
}

sub B::PVOP::toxml {
	my ($op, $level) = @_;
	print indent($level), "<pvop ", B::ToXML::get_attrs($op, 'pv'), "/>\n";
}

sub B::LOOP::toxml {
	my ($op, $level) = @_;

	print indent($level), "<loop ", B::ToXML::get_attrs($op), ">\n";

	my $kid = $op->first;
	do {
		$kid->toxml( $level + 1 );
	} while ( $kid = $kid->sibling and $$kid );

	print indent($level), "</loop>\n";
}

sub B::COP::toxml {
	my ($op, $level) = @_;

	print indent($level), "<cop ",
		B::ToXML::get_attrs( $op, qw( name label file line cop_seq flags
		private )), 
			"/>\n";
}

sub B::LOGOP::toxml {
	my ($op, $level) = @_;

	print indent($level), "<logop ", B::ToXML::get_attrs( $op ), qq|other="|, 
		$op->other->seq, qq|">\n|;

	my $kid = $op->first;
	do {
		$kid->toxml( $level + 1 );
	} while ( $kid = $kid->sibling and $$kid );

	print indent( $level ), "</logop>\n";
}

sub B::PV::toxml {
	my ($op, $level) = @_;
	my $pv = encode($op->PV) || '';
	print indent($level), qq|<pv>$pv</pv>\n|;
}

sub B::AV::toxml {
	my ($op, $level) = @_;
	print indent($level), "<av ", B::ToXML::get_attrs($op, 'fill'), "/>\n";
}

sub B::GV::toxml {
	my ($op, $level) = @_;

	my $stash = $op->STASH->NAME;
	$stash = encode($stash eq 'name' ? '' : "${stash}::");
	my $name = encode($op->SAFENAME);
	print indent($level), qq|<gv name="$name" stash="$stash" />\n|;
}

sub B::IV::toxml {
	my ($op, $level) = @_;
	my $val = encode($op->int_value);
	print indent($level), qq|<iv value="$val" />\n|;
}

sub B::NV::toxml {
	my ($op, $level) = @_;
	my $val = encode($op->NV);
	print indent($level), qq|<nv value="$val" />\n|;
}

sub B::RV::toxml {
	my ($op, $level) = @_;
	my $val = encode($op->RV);
	print indent($level), qq|<rv value="$val" />\n|;
}

sub B::NULL::toxml {
	my ($op, $level) = @_;
	#	print indent($level), "<null ", B::ToXML::get_attrs($op), "/>\n";
	print indent($level), "<null />\n";
}

sub B::SPECIAL::toxml {
	my ($op, $level) = @_;
	print indent($level), "<special ", B::ToXML::get_attrs($op), ">\n", 
		encode($specialsv_name[$$op]), "</special>\n";
}

sub indent {
	my $level = shift || 0;
	return "    " x $level;
}

sub get_attrs {
	my ($op, @fields) = @_;
	@fields = qw( name desc flags targ ) unless @fields;

	my $id = $op->seq;
	my $attr = qq|id="$id" |;
	foreach ( @fields ) {

		my $val = $op->$_();
		next unless defined $val;
		$val = encode( $val );
		$attr .= qq|$_="$val" |;
	}

	return $attr;
}

sub B::SV::toxml {
	warn "No SV!\n";
}

sub B::PVMG::toxml {
	warn "No PVMG!\n";
}

sub B::MAGIC::toxml {
	warn "No MAGIC!\n";
}

sub B::PVLV::toxml {
	warn "No PVLV!\n";
}

sub B::BM::toxml {
	warn "No BM!\n";
}

sub B::IO::toxml {
	warn "No IO!\n";
}

sub B::CV::toxml {
	warn "No CV!\n";
}

sub B::HV::toxml {
	warn "No HV!\n";
}

sub B::PMOP::toxml {
	warn "No PMOP!\n";
}

sub B::CVOP::toxml {
	warn "No CVOP!\n";
}

1;
